@@ -1,6 +1,6 @@
 /**
  * perl-libxml-mm.c
- * $Id: perl-libxml-mm.c 709 2008-01-29 21:01:32Z pajas $
+ * $Id: perl-libxml-mm.c 729 2008-10-30 13:16:23Z pajas $
  *
  * Basic concept:
  * perl varies in the implementation of UTF8 handling. this header (together
@@ -86,6 +86,22 @@ PmmNodeTypeName( xmlNodePtr elem ){
     return "";
 }
 
+/*
+ * free a hash table
+ */
+void
+PmmFreeHashTable(xmlHashTablePtr table)
+{
+	if( xmlHashSize(table) > 0 ) {
+		warn("PmmFreeHashTable: not empty\n");
+		/* PmmDumpRegistry(table); */
+	}
+	/*	warn("Freeing table %p with %d elements in\n", table, xmlHashSize(table)); */
+	xmlHashFree(table, NULL);
+}
+
+#ifdef XML_LIBXML_THREADS
+
 /*
  * registry of all current proxy nodes
  *
@@ -93,28 +109,103 @@ PmmNodeTypeName( xmlNodePtr elem ){
  * to this registry via XML::LibXML::__proxy_registry
  *
  */
-extern ProxyNodePtr PROXY_NODE_REGISTRY;
+extern SV* PROXY_NODE_REGISTRY_MUTEX;
+
+/* Utility method used by PmmDumpRegistry */
+void PmmRegistryDumpHashScanner(void * payload, void * data, xmlChar * name)
+{
+	LocalProxyNodePtr lp = (LocalProxyNodePtr) payload;
+	ProxyNodePtr node = (ProxyNodePtr) lp->proxy;
+	const char * CLASS = PmmNodeTypeName( PmmNODE(node) );
+	warn("%s=%p with %d references (%d perl)\n",CLASS,node,PmmREFCNT(node),lp->count);
+}
+
+/*
+ * dump the current thread's node registry to STDERR
+ */
+void
+PmmDumpRegistry(xmlHashTablePtr r)
+{
+	if( r )
+	{
+		SvLOCK(PROXY_NODE_REGISTRY_MUTEX);
+		warn("%d total nodes\n", xmlHashSize(r));
+		xmlHashScan(r, PmmRegistryDumpHashScanner, NULL);
+		SvUNLOCK(PROXY_NODE_REGISTRY_MUTEX);
+	}
+}
 
 /*
  * returns the address of the proxy registry
  */
-ProxyNodePtr*
+xmlHashTablePtr*
 PmmProxyNodeRegistryPtr(ProxyNodePtr proxy)
 {
-    return &PROXY_NODE_REGISTRY;
+	croak("PmmProxyNodeRegistryPtr: TODO!\n");
+	return NULL;
+	/*   return &PmmREGISTRY; */
 }
 
+/*
+ * efficiently generate a string representation of the given pointer
+ */
+xmlChar *
+PmmRegistryName(void * ptr)
+{
+	unsigned long int v = (unsigned long int) ptr;
+	int HASH_NAME_SIZE = sizeof(void *) + ceil(sizeof(void *)/8);
+	xmlChar * name;
+	int i;
+
+	name = (xmlChar *) safemalloc(HASH_NAME_SIZE+1);
+
+	for(i = 0; i < HASH_NAME_SIZE; ++i)
+	{
+		name[i] = (xmlChar) (128 | v);
+		v >>= 7;
+	}
+	name[HASH_NAME_SIZE] = '\0';
+
+	return name;
+}
+
+/*
+ * allocate and return a new LocalProxyNode structure
+ */
+LocalProxyNodePtr
+PmmNewLocalProxyNode(ProxyNodePtr proxy)
+{
+	LocalProxyNodePtr lp;
+    Newc(0, lp, 1, LocalProxyNode, LocalProxyNode);
+	lp->proxy = proxy;
+	lp->count = 0;
+	return lp;
+}
 
 /*
  * @proxy: proxy node to register
  *
  * adds a proxy node to the proxy node registry
  */
-void
+LocalProxyNodePtr
 PmmRegisterProxyNode(ProxyNodePtr proxy)
 {
-    proxy->_registry = PROXY_NODE_REGISTRY;
-    PROXY_NODE_REGISTRY = proxy;
+	xmlChar * name = PmmRegistryName( proxy );
+	LocalProxyNodePtr lp = PmmNewLocalProxyNode( proxy );
+        /* warn("LibXML registers proxy node with %p\n",PmmREGISTRY); */
+	SvLOCK(PROXY_NODE_REGISTRY_MUTEX);
+	if( xmlHashAddEntry(PmmREGISTRY, name, lp) )
+		croak("PmmRegisterProxyNode: error adding node to hash, hash size is %d\n",xmlHashSize(PmmREGISTRY));
+	SvUNLOCK(PROXY_NODE_REGISTRY_MUTEX);
+	Safefree(name);
+	return lp;
+}
+
+/* utility method for PmmUnregisterProxyNode */
+static inline void
+PmmRegistryHashDeallocator(void *payload, xmlChar *name)
+{
+	Safefree((LocalProxyNodePtr) payload);
 }
 
 /*
@@ -125,23 +216,65 @@ PmmRegisterProxyNode(ProxyNodePtr proxy)
 void
 PmmUnregisterProxyNode(ProxyNodePtr proxy)
 {
-    ProxyNodePtr cur = PROXY_NODE_REGISTRY;
-    if( PROXY_NODE_REGISTRY == proxy ) {
-        PROXY_NODE_REGISTRY = proxy->_registry;
-    }
-    else if (cur) {
-        while(cur->_registry != NULL)
-        {
-            if( cur->_registry == proxy )
-            {
-                cur->_registry = proxy->_registry;
-                break;
-            }
-            cur = cur->_registry;
-        }
-    } else {
-      warn("XML::LibXML: unregistering node, while no nodes have been registered?\n");
-    }
+	xmlChar * name = PmmRegistryName( proxy );
+        /* warn("LibXML unregistering proxy node with %p\n",PmmREGISTRY); */
+	SvLOCK(PROXY_NODE_REGISTRY_MUTEX);
+	if( xmlHashRemoveEntry(PmmREGISTRY, name, PmmRegistryHashDeallocator) )
+            croak("PmmUnregisterProxyNode: error removing node from hash\n");
+	Safefree(name);
+	SvUNLOCK(PROXY_NODE_REGISTRY_MUTEX);
+}
+
+/*
+ * lookup a LocalProxyNode in the registry
+ */
+LocalProxyNodePtr
+PmmRegistryLookup(ProxyNodePtr proxy)
+{
+	xmlChar * name = PmmRegistryName( proxy );
+	LocalProxyNodePtr lp = xmlHashLookup(PmmREGISTRY, name);
+	Safefree(name);
+	return lp;
+}
+
+/*
+ * increment the local refcount for proxy
+ */
+void
+PmmRegistryREFCNT_inc(ProxyNodePtr proxy)
+{
+  /* warn("Registry inc\n"); */
+	LocalProxyNodePtr lp = PmmRegistryLookup( proxy );
+	if( lp )
+		lp->count++;
+	else
+		PmmRegisterProxyNode( proxy )->count++;
+}
+
+/*
+ * decrement the local refcount for proxy and remove the local pointer if zero
+ */
+void
+PmmRegistryREFCNT_dec(ProxyNodePtr proxy)
+{
+  /* warn("Registry dec\n"); */
+	LocalProxyNodePtr lp = PmmRegistryLookup(proxy);
+	if( lp && --(lp->count) == 0 )
+		PmmUnregisterProxyNode(proxy);
+}
+
+/*
+ * internal, used by PmmCloneProxyNodes
+ */
+void *
+PmmRegistryHashCopier(void *payload, xmlChar *name)
+{
+	ProxyNodePtr proxy = ((LocalProxyNodePtr) payload)->proxy;
+	LocalProxyNodePtr lp;
+	Newc(0, lp, 1, LocalProxyNode, LocalProxyNode);
+	memcpy(lp, payload, sizeof(LocalProxyNode));
+	PmmREFCNT_inc(proxy);
+	return lp;
 }
 
 /*
@@ -150,12 +283,12 @@ PmmUnregisterProxyNode(ProxyNodePtr proxy)
 void
 PmmCloneProxyNodes()
 {
-    ProxyNodePtr cur = PROXY_NODE_REGISTRY;
-    while(cur != NULL)
-    {
-        PmmREFCNT_inc(cur);
-        cur = cur->_registry;
-    }
+	SV *sv_reg = get_sv("XML::LibXML::__PROXY_NODE_REGISTRY",0);
+	xmlHashTablePtr reg_copy;
+	SvLOCK(PROXY_NODE_REGISTRY_MUTEX);
+	reg_copy = xmlHashCopy(PmmREGISTRY, PmmRegistryHashCopier);
+	SvIV_set(SvRV(sv_reg), PTR2IV(reg_copy));
+	SvUNLOCK(PROXY_NODE_REGISTRY_MUTEX);
 }
 
 /*
@@ -164,16 +297,11 @@ PmmCloneProxyNodes()
 int
 PmmProxyNodeRegistrySize()
 {
-    int i = 0;
-    ProxyNodePtr cur = PROXY_NODE_REGISTRY;
-    while(cur != NULL)
-    {
-        ++i;
-        cur = cur->_registry;
-    }
-    return i;
+	return xmlHashSize(PmmREGISTRY);
 }
 
+#endif /* end of XML_LIBXML_THREADS */
+
 /* creates a new proxy node from a given node. this function is aware
  * about the fact that a node may already has a proxy structure.
  */
@@ -188,16 +316,13 @@ PmmNewNode(xmlNodePtr node)
     }
 
     if ( node->_private == NULL ) {
-        /* proxy = (ProxyNodePtr)malloc(sizeof(struct _ProxyNode));  */
-        Newc(0, proxy, 1, ProxyNode, ProxyNode);
+	proxy = (ProxyNodePtr)xmlMalloc(sizeof(struct _ProxyNode));
         if (proxy != NULL) {
             proxy->node  = node;
             proxy->owner   = NULL;
             proxy->count   = 0;
             proxy->encoding= 0;
-            proxy->_registry = NULL;
             node->_private = (void*) proxy;
-            PmmRegisterProxyNode(proxy);
         }
     }
     else {
@@ -290,7 +415,7 @@ PmmREFCNT_dec( ProxyNodePtr node )
         retval = PmmREFCNT(node)--;
 	/* fprintf(stderr, "REFCNT on 0x%08.8X decremented to %d\n", node, PmmREFCNT(node)); */
         if ( PmmREFCNT(node) < 0 )
-            warn( "PmmREFCNT_dec: REFCNT decremented below 0!" );
+            warn( "PmmREFCNT_dec: REFCNT decremented below 0 for %p!", node );
         if ( PmmREFCNT(node) <= 0 ) {
             xs_warn( "PmmREFCNT_dec: NODE DELETION\n" );
 
@@ -325,9 +450,10 @@ PmmREFCNT_dec( ProxyNodePtr node )
                 
                 PmmFreeNode( libnode );
             }
-            PmmUnregisterProxyNode(node);
-            Safefree( node );
-            /* free( node ); */
+			else {
+				xs_warn( "PmmREFCNT_dec:   NO OWNER\n" );
+			}
+            xmlFree( node );
         }
     }
     else {
@@ -357,6 +483,10 @@ PmmNodeToSv( xmlNodePtr node, ProxyNodePtr owner )
     const char * CLASS = "XML::LibXML::Node";
 
     if ( node != NULL ) {
+#ifdef XML_LIBXML_THREADS
+      if( PmmUSEREGISTRY )
+		SvLOCK(PROXY_NODE_REGISTRY_MUTEX);
+#endif
         /* find out about the class */
         CLASS = PmmNodeTypeName( node );
         xs_warn("PmmNodeToSv: return new perl node of class:\n");
@@ -380,12 +510,16 @@ PmmNodeToSv( xmlNodePtr node, ProxyNodePtr owner )
                 }
             }
             else {
-                xs_warn("PmmNodeToSv:   proxy creation failed!\n");
+                croak("XML::LibXML: failed to create a proxy node (out of memory?)\n");
             }
         }
 
         retval = NEWSV(0,0);
         sv_setref_pv( retval, CLASS, (void*)dfProxy );
+#ifdef XML_LIBXML_THREADS
+	if( PmmUSEREGISTRY )
+	    PmmRegistryREFCNT_inc(dfProxy);
+#endif
         PmmREFCNT_inc(dfProxy); 
         /* fprintf(stderr, "REFCNT incremented on node: 0x%08.8X\n", dfProxy); */
 
@@ -400,6 +534,10 @@ PmmNodeToSv( xmlNodePtr node, ProxyNodePtr owner )
         default:
             break;
         }
+#ifdef XML_LIBXML_THREADS
+      if( PmmUSEREGISTRY )
+		SvUNLOCK(PROXY_NODE_REGISTRY_MUTEX);
+#endif
     }
     else {
         xs_warn( "PmmNodeToSv: no node found!\n" );
@@ -705,7 +843,7 @@ PmmContextREFCNT_dec( ProxyNodePtr node )
     int retval = 0;
     if ( node != NULL ) {
         retval = PmmREFCNT(node)--;
-	/* fprintf(stderr, "REFCNT on context 0x%08.8X decremented to %d\n", node, PmmREFCNT(node)); */
+	/* fprintf(stderr, "REFCNT on context %p decremented to %d\n", node, PmmREFCNT(node)); */
         if ( PmmREFCNT(node) <= 0 ) {
             xs_warn( "PmmContextREFCNT_dec: NODE DELETION\n" );
             libnode = (xmlParserCtxtPtr)PmmNODE( node );
@@ -782,16 +920,52 @@ PmmSvContext( SV * scalar )
 xmlChar*
 PmmFastEncodeString( int charset,
                      const xmlChar *string,
-                     const xmlChar *encoding ) 
+                     const xmlChar *encoding,
+                     STRLEN len ) 
 {
     xmlCharEncodingHandlerPtr coder = NULL;
     xmlChar *retval = NULL;
     xmlBufferPtr in = NULL, out = NULL;
 
+    int i;
+    /* first check that the input is not ascii */
+    /* since we do not want to recode ascii as, say, UTF-16 */
+    if (len<0) len=xmlStrlen(string);
+    for (i=0; i<len; i++) {
+        if(!string[i] || string[i] & 0x80) {
+            break;
+        }
+    }
+    if (i>=len) return xmlStrdup( string );
+    xs_warn("PmmFastEncodeString: string is non-ascii\n");
+
+    if ( charset == XML_CHAR_ENCODING_ERROR){
+        if (xmlStrcmp(encoding,(const xmlChar*)"UTF-16LE")==0) {
+            charset = XML_CHAR_ENCODING_UTF16LE;
+        } else if (xmlStrcmp(encoding,(const xmlChar*) "UTF-16BE")==0) {
+            charset = XML_CHAR_ENCODING_UTF16BE;
+        }
+    }
     if ( charset == XML_CHAR_ENCODING_UTF8 ) {
         /* warn("use UTF8 for encoding ... %s ", string); */
         return xmlStrdup( string );
     }
+    else if ( charset == XML_CHAR_ENCODING_UTF16LE || charset == XML_CHAR_ENCODING_UTF16BE ){
+        /* detect and strip BOM, if any */
+        if (len>=2 && (char)string[0]=='\xFE' && (char)string[1]=='\xFF') {
+            xs_warn("detected BE BOM\n");
+            string += 2;
+            len    -= 2;
+            coder = xmlGetCharEncodingHandler( XML_CHAR_ENCODING_UTF16BE );
+        } else if (len>=2 && (char)string[0]=='\xFF' && (char)string[1]=='\xFE') {
+            xs_warn("detected LE BOM\n");
+            string += 2;
+            len    -= 2;
+            coder = xmlGetCharEncodingHandler( XML_CHAR_ENCODING_UTF16LE );
+        } else {
+            coder= xmlGetCharEncodingHandler( charset );
+        }
+    }
     else if ( charset == XML_CHAR_ENCODING_ERROR ){
         /* warn("no standard encoding %s\n", encoding); */
         coder =xmlFindCharEncodingHandler( (const char *)encoding );
@@ -806,9 +980,8 @@ PmmFastEncodeString( int charset,
 
     if ( coder != NULL ) {
         xs_warn("PmmFastEncodeString: coding machine found \n");
-        in    = xmlBufferCreate();
+        in    = xmlBufferCreateStatic((void*)string, len);
         out   = xmlBufferCreate();
-        xmlBufferCCat( in, (const char *) string );
         if ( xmlCharEncInFunc( coder, out, in ) >= 0 ) {
             retval = xmlStrdup( out->content );
             /* warn( "encoded string is %s" , retval); */
@@ -827,11 +1000,21 @@ PmmFastEncodeString( int charset,
 xmlChar*
 PmmFastDecodeString( int charset,
                      const xmlChar *string,
-                     const xmlChar *encoding) 
+                     const xmlChar *encoding,
+                     STRLEN* len ) 
 {
     xmlCharEncodingHandlerPtr coder = NULL;
     xmlChar *retval = NULL;
     xmlBufferPtr in = NULL, out = NULL;
+    if (len==NULL) return NULL;
+    *len = 0;
+    if ( charset == XML_CHAR_ENCODING_ERROR){
+        if (xmlStrcmp(encoding,(const xmlChar*)"UTF-16LE")==0) {
+            charset = XML_CHAR_ENCODING_UTF16LE;
+        } else if (xmlStrcmp(encoding,(const xmlChar*) "UTF-16BE")==0) {
+            charset = XML_CHAR_ENCODING_UTF16BE;
+        }
+    }
 
     if ( charset == XML_CHAR_ENCODING_UTF8 ) {
         return xmlStrdup( string );
@@ -840,7 +1023,7 @@ PmmFastDecodeString( int charset,
         coder = xmlFindCharEncodingHandler( (const char *) encoding );
     }
     else if ( charset == XML_CHAR_ENCODING_NONE ){
-        xs_warn("PmmFastDecodeString: no encoding found\n");
+        warn("PmmFastDecodeString: no encoding found\n");
     }
     else {
         coder= xmlGetCharEncodingHandler( charset );
@@ -848,15 +1031,14 @@ PmmFastDecodeString( int charset,
 
     if ( coder != NULL ) {
         /* warn( "do encoding %s", string ); */
-        in  = xmlBufferCreate();
+        in  = xmlBufferCreateStatic((void*)string,xmlStrlen(string));
         out = xmlBufferCreate();
-        
-        xmlBufferCat( in, string );        
         if ( xmlCharEncOutFunc( coder, out, in ) >= 0 ) {
-	  retval = xmlCharStrndup((const char *)xmlBufferContent(out), xmlBufferLength(out));
+          *len = xmlBufferLength(out);
+	  retval = xmlStrndup(xmlBufferContent(out), *len);
         }
         else {
-            xs_warn("PmmFastEncodeString: decoding error\n");
+            /* xs_warn("PmmFastEncodeString: decoding error\n"); */
         }
         
         xmlBufferFree( in );
@@ -871,7 +1053,7 @@ PmmFastDecodeString( int charset,
  * while the encodig has the name of the encoding of string
  **/ 
 xmlChar*
-PmmEncodeString( const char *encoding, const xmlChar *string ){
+PmmEncodeString( const char *encoding, const xmlChar *string, STRLEN len ){
     xmlCharEncoding enc;
     xmlChar *ret = NULL;
     
@@ -880,7 +1062,7 @@ PmmEncodeString( const char *encoding, const xmlChar *string ){
             xs_warn("PmmEncodeString: encoding to UTF-8 from:\n");
             xs_warn( encoding );
             enc = xmlParseCharEncoding( encoding );
-            ret = PmmFastEncodeString( enc, string, (const xmlChar *)encoding );
+            ret = PmmFastEncodeString( enc, string, (const xmlChar *)encoding,len);
         }
         else {
             /* if utf-8 is requested we do nothing */
@@ -890,31 +1072,6 @@ PmmEncodeString( const char *encoding, const xmlChar *string ){
     return ret;
 }
 
-/**
- * decodeString returns an $encoding encoded string.
- * while string is an UTF-8 encoded string and 
- * encoding is the coding name
- **/
-char*
-PmmDecodeString( const char *encoding, const xmlChar *string){
-    char *ret=NULL;
-    xmlCharEncoding enc;
-
-    if ( string != NULL ) {
-        xs_warn( "PmmDecodeString called\n" );
-        if( encoding != NULL ) {
-            enc = xmlParseCharEncoding( encoding );
-            ret = (char*)PmmFastDecodeString( enc, string, (const xmlChar*)encoding );
-            xs_warn( "PmmDecodeString done\n" );
-        }
-        else {
-            ret = (char*)xmlStrdup(string);
-        }
-    }
-    return ret;
-}
-
-
 SV*
 C2Sv( const xmlChar *string, const xmlChar *encoding )
 {
@@ -969,7 +1126,7 @@ Sv2C( SV* scalar, const xmlChar *encoding )
             if ( encoding != NULL ) {        
 #endif
                 xs_warn( "SV2C:   domEncodeString!\n" );
-                ts= PmmEncodeString( (const char *)encoding, string );
+                ts= PmmEncodeString( (const char *)encoding, string, len );
                 xs_warn( "SV2C:   done encoding!\n" );
                 if ( string != NULL ) {
                     xmlFree(string);
@@ -1007,23 +1164,16 @@ nodeC2Sv( const xmlChar * string,  xmlNodePtr refnode )
                 PmmNodeEncoding(real_doc) = XML_CHAR_ENCODING_UTF8;
             }
 
-            decoded = PmmFastDecodeString( PmmNodeEncoding(real_doc) ,
+            decoded = PmmFastDecodeString( PmmNodeEncoding(real_doc),
                                            (const xmlChar *)string,
-                                           (const xmlChar*)real_doc->encoding);
+                                           (const xmlChar *)real_doc->encoding, 
+                                           &len );
 
             xs_warn( "push decoded string into SV" );
-            len = xmlStrlen( decoded );
             retval = newSVpvn( (const char *)decoded, len );
             xmlFree( decoded );
 
             if ( PmmNodeEncoding( real_doc ) == XML_CHAR_ENCODING_UTF8 ) {
-                /* most probably true, since libxml2 always 
-                 * sets doc->charset to UTF8, see tree.c:
-                 *
-                 * The in memory encoding is always UTF8
-                 * This field will never change and would
-                 * be obsolete if not for binary compatibility.
-                 */
 #ifdef HAVE_UTF8
                 xs_warn("nodeC2Sv: set UTF8-SV-flag\n");
                 SvUTF8_on(retval);
@@ -1055,40 +1205,37 @@ nodeSv2C( SV * scalar, xmlNodePtr refnode )
             if ( scalar != NULL && scalar != &PL_sv_undef ) {
                 STRLEN len = 0;
                 char * t_pv =SvPV(scalar, len);
-                xmlChar* ts = NULL;
-                xmlChar* string = xmlStrdup((xmlChar*)t_pv);
-                if ( xmlStrlen(string) > 0 ) {
+                xmlChar* string = NULL;
+                if ( t_pv && len > 0 ) {
                     xs_warn( "nodeSv2C:   no undefs\n" );
 #ifdef HAVE_UTF8
                     xs_warn( "nodeSv2C:   use UTF8\n" );
-                    if( !DO_UTF8(scalar) && real_dom != NULL && real_dom->encoding != NULL )
-#else
-                    if ( real_dom != NULL && real_dom->encoding != NULL )
+                    if( !DO_UTF8(scalar) ) {
 #endif
-		    {
                         xs_warn( "nodeSv2C:     domEncodeString!\n" );
-                        /*  if ( string == NULL || *string == 0 ) warn("string is empty" ); */
                         /* The following statement is to handle bad
                            values set by XML::LibXSLT */
                         if ( PmmNodeEncoding(real_dom) == XML_CHAR_ENCODING_NONE ) {
                             PmmNodeEncoding(real_dom) = XML_CHAR_ENCODING_UTF8;
                         }
-
-                        ts= PmmFastEncodeString( PmmNodeEncoding(real_dom),
-                                                 string,
-                                                 (const xmlChar*)real_dom->encoding );
+                        /* the following allocates a new string (by xmlStrdup if no conversion is done) */
+                        string= PmmFastEncodeString( PmmNodeEncoding(real_dom),
+                                                 (xmlChar*) t_pv,
+                                                 (const xmlChar*)real_dom->encoding,
+                                                 len);
                         xs_warn( "nodeSv2C:     done!\n" );
-                        if ( string != NULL ) {
-                            xmlFree(string);
-                        }
-                        string=ts;
-                    }
-                    else {
+#ifdef HAVE_UTF8
+                    } else {
                         xs_warn( "nodeSv2C:   no encoding set, use UTF8!\n" );
                     }
+#endif
+                } 
+                if (string==NULL) {
+                    return xmlStrndup((xmlChar*)t_pv,len);
+                } else {
+                    return string;
                 }
                 /* if ( string == NULL ) warn( "nodeSv2C:     string is NULL\n" ); */
-                return string;
             }
             else {
                 xs_warn( "nodeSv2C:   return NULL\n" );
